local FS = LibStub("AceAddon-3.0"):GetAddon("FaceShooter")

----------
--- FS ---
----------

-- Copies the src table while reusing the memory already allocated for dst table and skipping (static) functions
function FS:CopyTable(src, dst)
	
	-- clear the destination table so there not variables that are nil in the source table
	for akey, avalue in pairs(dst) do
		if type(avalue) == "table" then
			for bkey, bvalue in pairs(avalue) do
				if type(bvalue) == "table" then
					for ckey, cvalue in pairs(bvalue) do
						if type(cvalue) == "table" then
							for dkey, dvalue in pairs(cvalue) do
								if type(dvalue) ~= "function" then dst[akey][bkey][ckey][dkey] = nil
								end
							end
						elseif type(cvalue) ~= "function" then dst[akey][bkey][ckey] = nil
						end
					end
				elseif type(bvalue) ~= "function" then dst[akey][bkey] = nil
				end
			end
		elseif type(avalue) ~= "function" then dst[akey] = nil
		end
	end
	
	-- copy table by hand to avoid referencing and memory issues
	for akey, avalue in pairs(src) do
		if type(avalue) == "table" then
			for bkey, bvalue in pairs(avalue) do
				if type(bvalue) == "table" then
					for ckey, cvalue in pairs(bvalue) do
						if type(cvalue) == "table" then
							for dkey, dvalue in pairs(cvalue) do
								if type(dvalue) ~= "function" then dst[akey][bkey][ckey][dkey] = dvalue
								end
							end
						elseif type(cvalue) ~= "function" then dst[akey][bkey][ckey] = cvalue
						end
					end
				elseif type(bvalue) ~= "function" then dst[akey][bkey] = bvalue
				end
			end
		elseif type(avalue) ~= "function" then dst[akey] = avalue
		end
	end
end

-- Gets a table from the pool or creates a new one
function FS:GetTable()
	return table.remove(self.freeTables) or {}
end

-- Empties the given table and puts it in the pool
function FS:RecycleTable(tbl)
	for key in pairs(tbl) do
		tbl[key] = nil
	end
	table.insert(self.freeTables, tbl)
end

-- removes an existing shot object
function FS:RemoveButton(name)
	self[name]["mov"]:Hide()
	self:RecycleTable(self[name]["mov"])
	self[name]["mov"] = nil
	self[name]["btn"]:Hide()
	self:RecycleTable(self[name]["btn"])
	self[name]["btn"] = nil
	self[name]["cdf"]:Hide()
	self:RecycleTable(self[name]["cdf"])
	self[name]["cdf"] = nil
	self[name].resources = nil
	self:RecycleTable(self[name])
	self[name] = nil
end

-- simple function to round numbers
function FS:Round(num, digits)
	if num == nil then return 0 end
	if not tonumber(num) then return num end
	local mult = 10 ^ (digits or 0)
	return floor(num * mult + 0.5) / mult
end

-- called when a valid new target is acquired
function FS:AcquireTarget()

	self.tarGUID = UnitGUID("target")
	if self.state == "idle" then self.state = "waiting" end
	for i = 1, 2 do
		if self.db.profile["button"..i]["enabled"] then self["button"..i]:Show() end
	end
	for name, module in FS:IterateModules() do
		if module:IsEnabled() and module.AcquireTarget then module:AcquireTarget() end
	end
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("UNIT_SPELLCAST_SENT")
	self:RegisterEvent("UNIT_SPELLCAST_START")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("SPELL_UPDATE_COOLDOWN")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self.resources:GetDebuffs()
	
	self:Scheduler()
end

-- called when target is lost or not valid
function FS:LooseTarget()
	self.tarGUID = nil
	self.state = "idle"
	for i = 1, 2 do
		if self.db.profile["button"..i]["enabled"] then self["button"..i]:Hide() end
	end
	for name, module in FS:IterateModules() do
		if module:IsEnabled() and module.LooseTarget then module:LooseTarget() end
	end
	self:UnregisterEvent("UNIT_AURA")
	self:UnregisterEvent("UNIT_SPELLCAST_SENT")
	self:UnregisterEvent("UNIT_SPELLCAST_START")
	self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:UnregisterEvent("SPELL_UPDATE_COOLDOWN")
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")

end

-- calculate the average of a given statType over a given amount of datapoints
function FS:GetAverage(statType, data, dataPoints)
	dataPoints = dataPoints or 10
	if not self[statType] then self[statType] = {} end
	table.insert(self[statType], 1, data)
	table.remove(self[statType], dataPoints)
	local sum, num = 0, 0
	for key, val in pairs(self[statType]) do
		sum = sum + val
		num = num + 1
	end
	return (sum / num)
end

-- determine which options need to be activated or deactivated
function FS:CheckForDisabled()
	if self.db.profile.ba.suggestBA == "never" then FS.options.args.ba.args.saveFocusBA.disabled = true
	else FS.options.args.ba.args.saveFocusBA.disabled = false
	end
end

-- check for interrupts
function FS:CheckForInterrups()
	local spellName, rank, displayName, icon, startTime, endTime
	spellName, _, _, _, startTime, endTime = UnitChannelInfo("player")
	if spellName == nil then 
		spellName, rank, displayName, icon, startTime, endTime = UnitCastingInfo("player")
	end
	if spellName then self.castEndsAt = endTime / 1000 end
	-- spellNum table should be empty while not casting anything
	if self.state ~= "castStarted" then
		if next(self.spellNum) then
			for key, val in pairs(self.spellNum) do
				if val == self.spellName[0] then
					self.spellName[0] = nil
					self.spellSentAt[0] = nil
					self.spellStartedAt[0] = nil
					self.spellEndedAt[0] = nil
				end
				if val == GetSpellInfo(self.spellID["SS"]) then self.resources.focusGains["SS"] = self.resources.focusGains["SS"] - 1
				elseif val == GetSpellInfo(self.spellID["CoS"]) then self.resources.focusGains["CoS"] = self.resources.focusGains["CoS"] - 1
				elseif val == GetSpellInfo(self.spellID["FS"]) then self.resources.focusGains["FS"] = self.resources.focusGains["FS"] - 1
				end
				self.spellNum[key] = nil
			end
			if self.resources.focusGains["SS"] < 0 then self.resources.focusGains["SS"] = 0 end
			if self.resources.focusGains["CoS"] < 0 then self.resources.focusGains["CoS"] = 0 end
			if self.resources.focusGains["FS"] < 0 then self.resources.focusGains["FS"] = 0 end
		end
	
	-- if in castStarted state but no spell casting
	elseif not spellName then
		local now = GetTime()
		if now < self.castEndsAt or now > (self.castEndsAt + self.avrgLatency) then
			for key, val in pairs(self.spellNum) do	self.spellNum[key] = nil end
			self.spellName[0] = nil
			self.spellName[1] = nil
			self.spellSentAt[0] = nil
			self.spellSentAt[1] = nil
			self.spellStartedAt[0] = nil
			self.spellStartedAt[1] = nil
			self.spellEndedAt[0] = nil
			self.spellEndedAt[1] = nil
			self.resources.focusGains["SS"] = 0
			self.resources.focusGains["CoS"] = 0
			self.resources.focusGains["FS"] = 0
			self.state = "waiting"
		end
	end
end

-- checks if player is in a raid
function FS:IsInRaid()
	return IsInRaid("player")
end

-- checks if player is in a party
function FS:IsInParty()
	if not IsInRaid("player") and (GetNumGroupMembers() > 0) then
		return true
	else
		return false
	end
end

-- checks if player is solo
function FS:IsSolo()
	if GetNumGroupMembers() > 0 then
		return false
	else
		return true
	end
end

-- checks if player is mounted or in a vehicle
function FS:IsMounted()
	local TT = FS.resources.buffs["TT"] ~= 0
	local FWW = FS.resources.buffs["FWW"] ~= 0

	if (TT or FWW) then return false end
	if (IsMounted() or UnitInVehicle("player")) then
		return true
	else
		return false
	end
end

-- checks if target is a valid target according to settings
function FS:IsLegitTarget()
	
	-- if player has no target, target is friendly, target is dead or a player character
	if UnitName("target") == nil or UnitHealth("target") == 0 or UnitIsPlayer("target") or UnitIsFriend("player","target") then
		return false
	-- if player is in raid and showInRaid is false or player in party and showInParty is false etc.
	elseif (self:IsInRaid() and not self.db.profile.showInRaid) or (self:IsInParty() and not self.db.profile.showInParty) or (self:IsSolo() and not self.db.profile.showWhenSolo) then
		return false
	-- if player is flagged as mounted
	elseif self.isMounted then
		return false
	else
		return true
	end
end

function FS:FindKillShotSlot(slot)
	if not slot then slot = 0 end
	
	--FS:Print("Start FindKillShotSlot:"..slot)
	if slot > 0 then
		--FS:Print("Start direct check")
		local _, SpellID = GetActionInfo(slot)
		local Texture = GetActionTexture(slot)
		--FS:Print(Texture)
		if SpellID == 157708 or Texture == "Interface\\Icons\\Ability_Hunter_Assassinate2" then 
			FS.ksslot = slot
			--FS:Print("Direct: Slot " .. FS.ksslot)
			return
		else -- Check it is still where it used to be...
			local _, SpellID = GetActionInfo(FS.ksslot)
			local Texture = GetActionTexture(FS.ksslot)
			if SpellID == 157708 or Texture == "Interface\\Icons\\Ability_Hunter_Assassinate2" then 
				--FS:Print("Direct: Slot unchanged " .. FS.ksslot)
				return
			end
		end
	end

	 -- Full Search for Kill Shot through all action bars
	FS.ksslot = 0
	--FS:Print("Kill Shot Search")
	for slot = 1, 120 do --Really don't like this, is inefficient
		local _, SpellID = GetActionInfo(slot)
		local Texture = GetActionTexture(slot)
		if SpellID == 157708 or Texture == "Interface\\Icons\\Ability_Hunter_Assassinate2" then -- Don't know why this isn't 53351?
			FS.ksslot = slot
			--FS:Print("Search: Slot " .. FS.ksslot)
			break
		end
	end
	
	--Interface\Icons\Ability_Hunter_Assassinate2  (icon from GetSpellInfo)

	--if FS.ksslot == 0 then
	--	FS:Print("Faceshooter needs Kill Shot on one of your Action bars otherwise it cannot suggest Kill Shot...")
	--	FS:Print("Open your spell book and drag Kill Shot onto one of your bars now")
	--end
end

-- reads the talents and returns the spec
function FS:GetTalents()
	-- Find out which talents are available to the player
	local activeSpec = GetActiveSpecGroup();
	local specialization = GetSpecialization()

	self.talents = {}
	local _, _, _, curr = GetTalentInfo(4,1,activeSpec)
	self.talents["SF"] = curr -- Steady Focus
	local _, _, _, curr = GetTalentInfo(4,2,activeSpec)
	self.talents["DB"] = curr -- Dire Beast
	local _, _, _, curr = GetTalentInfo(4,3,activeSpec)
	self.talents["TotH"] = curr -- Thrill of the Hunt
	local _, _, _, curr = GetTalentInfo(5,1,activeSpec)
	self.talents["AMoC"] = curr -- A Murder of Crows
	local _, _, _, curr = GetTalentInfo(5,3,activeSpec)
	self.talents["ST"] = curr -- Stampede
	local _, _, _, curr = GetTalentInfo(6,1,activeSpec)
	self.talents["GT"] = curr -- Glaive Toss
	local _, _, _, curr = GetTalentInfo(6,2,activeSpec)
	self.talents["PS"] = curr -- Powershot
	local _, _, _, curr = GetTalentInfo(6,3,activeSpec)
	self.talents["BRG"] = curr -- Barrage
	local _, _, _, curr = GetTalentInfo(7,2,activeSpec)
	self.talents["FS"] = curr -- Focusing Shot
	local _, _, _, curr = GetTalentInfo(7,3,activeSpec) -- Either Adaptation or Lone Wolf
	if specialization == 1 then
		self.talents["LW"] = curr -- Lone Wolf
	end

	if specialization == 1 then
		return "BM"
	elseif specialization == 2 then
		return "MM"
	elseif specialization == 3 then
		return "SV"
	else
		return "none"	
	end
end

-- get the current gcd and return start and duration
function FS:GetGCD()

	local start, duration = GetSpellCooldown(FS.spellID["SS"])
	local gcd = start + duration
	local now = GetTime() 
	-- if cast has just been sent and the gcd is set but lower than 0.5 it's probably still the last casts GCD
	if self.state == "castSent" and gcd > 0 and gcd < now + 0.5 then
		self.gcd = now + 1
		return now, 1
	-- anytime else just use the values of GetSpellCooldown
	else
		self.gcd = gcd
		return start, duration
	end
	

end

-- returns how long the given cast will take
function FS:GetCastDelta(spell, haste )
	if spell == GetSpellInfo(self.spellID["SS"]) or spell == GetSpellInfo(self.spellID["CoS"]) then
		delta = 2 / haste
		--FS:Print("SS-CoS Cast delta: "..delta)
	elseif spell == GetSpellInfo(self.spellID["AimS"]) or spell == GetSpellInfo(self.spellID["FS"])then
		delta = 2.5 / haste
		--FS:Print("AimS - FS Cast delta: "..delta)
	elseif spell == GetSpellInfo(self.spellID["PS"]) then
		delta = 2.25 / haste
		--FS:Print("PS Cast delta: "..delta)
	elseif spell == GetSpellInfo(self.spellID["BRG"]) then
		delta = 3 / haste
		--FS:Print("BRG Cast delta: "..delta)
	else
		delta = 1
	end
	if delta < 1 then delta = 1	end
	return delta
end

----------------
-- resources ---
----------------

-- gets the resources as they are at the moment
local function GetResources(self)
	local now = GetTime()
	local focus = UnitPower("player")
	local rangedAttackSpeed = UnitRangedDamage("player")
	local start, duration = GetSpellCooldown(FS.spellID["SS"]) --Does this work if not MM?
	local gcd = start + duration
	local txt
	-- copy the current spellqueues
	FS:CopyTable(FS.spellName, self.spellName) -- CopyTable(source, destination)
	FS:CopyTable(FS.spellSentAt, self.spellSentAt)
	FS:CopyTable(FS.spellStartedAt, self.spellStartedAt)
	FS:CopyTable(FS.spellEndedAt, self.spellEndedAt)
	
	-- get haste, targets hp, focus and time
	self.focus = focus
	self.now = now
	self.tarHP = UnitHealth("target") / UnitHealthMax("target")
	self.haste = 1 + GetRangedHaste() / 100
	
	if FS.talents["FS"] then self.gains = 50
	else self.gains = 14 end

	-- get the CDs of the various used spells
	
	--Kill Shot
	start, duration = GetActionCooldown(FS.ksslot)
	--FS:Print("GCD:"..gcd.." KS:"..start.." :"..duration)
	if (start + duration) == gcd then self.cooldowns["KS"] = 0
	else self.cooldowns["KS"] = start + duration
	end
	
	start, duration = GetSpellCooldown(FS.spellID["CS"])
	if (start + duration) == gcd then self.cooldowns["CS"] = 0
	else self.cooldowns["CS"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["BA"])
	if (start + duration) == gcd then self.cooldowns["BA"] = 0
	else self.cooldowns["BA"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["ES"])
	if (start + duration) == gcd then self.cooldowns["ES"] = 0
	else self.cooldowns["ES"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["DB"])
	if (start + duration) == gcd then self.cooldowns["DB"] = 0
	else self.cooldowns["DB"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["AMoC"])
	if (start + duration) == gcd then self.cooldowns["AMoC"] = 0
	else self.cooldowns["AMoC"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["GT"])
	if (start + duration) == gcd then self.cooldowns["GT"] = 0
	else self.cooldowns["GT"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["PS"])
	if (start + duration) == gcd then self.cooldowns["PS"] = 0
	else self.cooldowns["PS"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["BRG"])
	if (start + duration) == gcd then self.cooldowns["BRG"] = 0
	else self.cooldowns["BRG"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["KC"])
	if (start + duration) == gcd then self.cooldowns["KC"] = 0
	else self.cooldowns["KC"] = start + duration
	end
	start, duration = GetSpellCooldown(FS.spellID["ET"])
	if (start + duration) == gcd then self.cooldowns["ET"] = 0
	else self.cooldowns["ET"] = start + duration
	end
	
	-- get buffs and debuffs
	self:GetBuffs()
	self:GetDebuffs()
end

-- sets the resources to the given values
local function SetResources(self, resources)
	FS:CopyTable(resources, self) --CopyTable(Source, Destination)
end

-- account for focus gains that happen after the cast
local function FocusGains(self)
	for key, val in pairs(self.focusGains) do
		if (key == "SS" or key == "CoS" or key == "FS") then 
			self.focus = self.focus + self.gains * val
			self.focusGains[key] = 0
		end
	end
	if FS.spec == "BM" and self.focus > 120 then self.focus = 120 end
	if FS.spec == "SV" and self.focus > 100 then self.focus = 100 end
	if FS.spec == "MM" and FS.perks then 
		if self.focus > 120 then self.focus = 120 end
	else
		if self.focus > 100 then self.focus = 100 end
	end

end

-- calculate the spellcosts
local function GetSpellCosts(self)
	
	if self.buffs["BW"] > self.now then -- BM only
		self.cost = {
			["KC"] = 20,
			["AS"] = 15,
			["AMoC"] = 15,
			["GT"] = 8,
			["PS"] = 8,
			["BRG"] = 30,
		}
		else self.cost = {
			["ES"] = 15,
			["CS"] = 35,
			["BA"] = 35,
			["KC"] = 40,
			["AS"] = 30,
			["AimS"] = 50,
			["AMoC"] = 30,
			["GT"] = 15,
			["PS"] = 15,
			["BRG"] = 60,
		}
	end

	if self.buffs["TotH"] > self.now then
		self.cost["AS"] = 10
		self.cost["AimS"] = 30
	else
		self.cost["AS"] = 30
		self.cost["AimS"] = 50
	end

	-- Does Bestial Wrath and TotH stack for Arcane Shot. Quick experiment shows yes
	if self.buffs["TotH"] > self.now and self.buffs["BW"] > self.now then
		self.cost["AS"] = 0
	end
end

-- get the buffs on yourself
local function GetBuffs(self)
	--for index = 0, 10 do
	--	local name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId = UnitBuff("player", index)
	--	if name == "Telaari Talbuk" then
	--		FS:Print ("Buff Name: ".. name)
	--		if icon ~= nil then	FS:Print ("Buff Icon: ".. icon)	else FS:Print("Icon: nil") end
	--		if count ~= nil then FS:Print ("Buff Count: ".. count)	else FS:Print("Count: nil") end
	--		if debuffType ~= nil then FS:Print ("Buff debuffType: ".. debuffType)	else FS:Print("debuffType: nil") end
	--		if duration ~= nil then FS:Print ("Buff duration: ".. duration)	else FS:Print("Duration: nil") end
	--		if expirationTime ~= nil then FS:Print ("Buff expirationTime: ".. expirationTime)	else FS:Print("expirationTime: nil") end
	--		if unitCaster ~= nil then FS:Print ("Buff unitCaster: ".. unitCaster)	else FS:Print("unitCaster: nil") end
	--		if spellId ~= nil then FS:Print ("Buff spellId: ".. spellId)	else FS:Print("spellId: nil") end
	--	end
	--end
	for buff, val in pairs(self.buffs) do
		local name, _, _, cnt, _, dur, expTime, caster = UnitBuff("player", GetSpellInfo(FS.spellID[buff]))
		if name then
			if name == "Frenzy" then
				self.buffs[buff] = expTime
			elseif name == "Telaari Talbuk" then
				self.buffs[buff] = 1
			elseif name == "Frostwolf War Wolf" then
				self.buffs[buff] = 1
			else 
				self.buffs[buff] = expTime
			end
		else
				self.buffs[buff] = 0
		end
	end
end

-- get the debuffs on target
local function GetDebuffs(self)
	--local name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId
	--local index
	--for index = 1, 10 do
	--	local name, rank, icon, count, debuffType, duration, expirationTime, unitCaster, isStealable, shouldConsolidate, spellId = UnitDebuff("target", index)
	--	if name ~= nil then
	--		FS:Print ("Buff Name: ".. name)
	--		if icon ~= nil then	FS:Print ("Buff Icon: ".. icon)	else FS:Print("Icon: nil") end
	--		if count ~= nil then FS:Print ("Buff Count: ".. count)	else FS:Print("Count: nil") end
	--		if debuffType ~= nil then FS:Print ("Buff debuffType: ".. debuffType)	else FS:Print("debuffType: nil") end
	--		if duration ~= nil then FS:Print ("Buff duration: ".. duration)	else FS:Print("Duration: nil") end
	--		if expirationTime ~= nil then FS:Print ("Buff expirationTime: ".. expirationTime)	else FS:Print("expirationTime: nil") end
	--		if unitCaster ~= nil then FS:Print ("Buff unitCaster: ".. unitCaster)	else FS:Print("unitCaster: nil") end
	--		if spellId ~= nil then FS:Print ("Buff spellId: ".. spellId)	else FS:Print("spellId: nil") end
	--	end
	--end
	
	local now, filter = GetTime()
	for debuff, val in pairs(self.debuffs) do
		if debuff == "AMoC" then filter = "PLAYER" else filter = nil end
		local name, _, _, cnt, _, dur, expTime, caster = UnitDebuff("target", GetSpellInfo(FS.spellID[debuff]), nil, filter)
		-- if spell has been found use the found duration
		if name then
			-- A Murder of Crows doesn't return expTime or Duration and is just set to "present"
			if name == GetSpellInfo(FS.spellID["AMoC"]) then
				self.debuffs[debuff][FS.tarGUID] = now + 10
			-- otherwise only consider our own debuffs
			elseif name == GetSpellInfo(FS.spellID["SpS"]) then
				self.debuffs[debuff][FS.tarGUID] = expTime
			elseif caster == "player" then
				self.debuffs[debuff][FS.tarGUID] = expTime
			else
				self.debuffs[debuff][FS.tarGUID] = 0
			end
		-- if spell with that name has just been cast assume it just hasn't reached the server yet
		elseif self.lastCast[debuff] and now < (self.lastCast[debuff] + 1 + FS.avrgLatency) then
			self.debuffs[debuff][FS.tarGUID] = now + 15
		-- otherwise set debuff duration to 0
		else
			self.debuffs[debuff][FS.tarGUID] = 0
		end
	end
end

-- when ooc purge remaining debuffs on everything thats not your target
local function PurgeDebuffs(self)
	for debuff, val in pairs(self.debuffs) do
		for id, expTime in pairs(val) do
			if id ~= FS.tarGUID then self.debuffs[debuff][id] = nil end
		end
	end
end

-- returns assumed changes to resources when casting given spell
-- CastSpell is called by Scheduler with either 'fs' as parameter or button number ("button1" or "button2")
local function CastSpell(self, parent)
	local now = GetTime()

	-- get the self.spell for which we're trying to find the outcome
	if self.spellName[0] then self.spell = self.spellName[0]
	elseif self.spellName[1] then self.spell = self.spellName[1]
	end

	-- calculate how long we have to wait after casting given spell
	local delta = FS:GetCastDelta(self.spell, self.haste)
	
	-- if a cast has been started use that starting time as base for the timer calculations
	if parent ~= "fs" then self.now = self.spellSentAt[1]
	elseif FS.state ~= "waiting" and self.spellName[0] then self.now = self.spellSentAt[0]
	elseif FS.state ~= "waiting" and self.spellName[1] then self.now = self.spellSentAt[1]
	elseif FS.state ~= "waiting" and self.spellName[2] then self.now = self.spellSentAt[2]
	else self.now = now -- In Waiting state
	end
	
	-- Calculate cooldowns, buffs and debuffs
	self:CalcTimers(parent)
	
	-- Calculate focus - not required after focus has been subtracted server side
	if parent ~= "fs" then self:CalcFocus(parent) --ie button2.resources, we must calc focus for the 2nd shot after being passed button1.resources
	elseif FS.state ~= "castStarted" and FS.state ~= "succeeded" then self:CalcFocus(parent)
	elseif FS.state == "castStarted" and (self.spell == GetSpellInfo(FS.spellID["AimS"])) then self:CalcFocus(parent)
	elseif self.spellName[0] then self:CalcFocus(parent)
	end
	
	-- if a spell has been queued next cast will be at the end of that spell + delta
	local timeToReg
	if parent ~= "fs" then
		self.now = self.spellEndedAt[1]
		timeToReg = self.spellEndedAt[1] - self.spellSentAt[1]
	elseif self.spellName[0] then
		self.now = self.spellEndedAt[0]
		timeToReg = self.spellEndedAt[0] - self.spellSentAt[0]
	elseif self.spellName[1] then
		self.now = self.spellEndedAt[1]
		timeToReg = self.spellEndedAt[1] - now
	elseif self.spellName[2] and self.spellEndedAt[2] > now then
		self.now = self.spellEndedAt[2]
		timeToReg = self.spellEndedAt[2] - now
	else
		self.now = now
		timeToReg = 0
	end

	-- account for lost or gained SF buffs
	local steady = GetSpellInfo(FS.spellID["SS"])
	local cobra = GetSpellInfo(FS.spellID["CoS"])
	if self.spec == "MM" then
		if self.spellName[0] == steady and self.spellName[1] == steady then self.buffs["SF"] = self.spellEndedAt[0] + 10 + FS.avrgLatency
		elseif self.spellName[1] == steady and self.spellName[2] == steady then self.buffs["SF"] = self.spellEndedAt[1] + 10 + FS.avrgLatency
		elseif self.spellName[2] == steady and self.spellName[3] == steady and self.spellEndedAt[2] > (now - 2) then self.buffs["SF"] = self.spellEndedAt[2] + 10 + FS.avrgLatency
		end
	else
		if self.spellName[0] == cobra and self.spellName[1] == cobra then self.buffs["SF"] = self.spellEndedAt[0] + 10 + FS.avrgLatency
		elseif self.spellName[1] == cobra and self.spellName[2] == cobra then self.buffs["SF"] = self.spellEndedAt[1] + 10 + FS.avrgLatency
		elseif self.spellName[2] == cobra and self.spellName[3] == cobra and self.spellEndedAt[2] > (now - 2) then self.buffs["SF"] = self.spellEndedAt[2] + 10 + FS.avrgLatency
		end
	end

	if FS.resources.buffs["SF"] >= GetTime() and self.buffs["SF"] < self.now then
		self.regenRate = 4
	elseif FS.resources.buffs["SF"] < GetTime() and self.buffs["SF"] >= self.now then
		self.regenRate = 6
	else
		self.regenRate = 4
	end
	
	self:CalcReg(timeToReg, self.regenRate)
end

-- calculate cooldowns and debuff durations when casting given spell
local function CalcTimers(self, parent)
	local spell = self.spell

	-- Cobra Shot
	if spell == GetSpellInfo(FS.spellID["CoS"]) then

	-- Explosive Shot
	elseif spell == GetSpellInfo(FS.spellID["ES"]) then
		if self.buffs["LnL"] > self.now then
			self.cooldowns["ES"] = 0
			self.lnl = true
		elseif parent ~= "fs" then
			-- trust the API for the fs frame in order to prevent setting a wrong CD after use of last LnL charge
			self.cooldowns["ES"] = self.now + 6
		end
	-- Kill Command
	elseif spell == GetSpellInfo(FS.spellID["KC"]) then
		self.cooldowns["KC"] = self.now + 6
	-- Chimaera Shot
	elseif spell == GetSpellInfo(FS.spellID["CS"]) then
		self.cooldowns["CS"] = self.now + 9
	-- Black Arrow
	elseif spell == GetSpellInfo(FS.spellID["BA"]) then
		self.cooldowns["BA"] = self.now + 24
		self.debuffs["BA"][FS.tarGUID] = self.now + 20
	-- Dire Beast
	elseif spell == GetSpellInfo(FS.spellID["DB"]) then
		self.cooldowns["DB"] = self.now + 30
	-- A Murder of Crows
	elseif spell == GetSpellInfo(FS.spellID["AMoC"]) then
		self.cooldowns["AMoC"] = self.now + 60
		self.debuffs["AMoC"][FS.tarGUID] = self.now + 15
	-- Glaive Toss
	elseif spell == GetSpellInfo(FS.spellID["GT"]) then
		self.cooldowns["GT"] = self.now + 15
	-- Powershot
	elseif spell == GetSpellInfo(FS.spellID["PS"]) then
		self.cooldowns["PS"] = self.now + 45
	-- Barrage
	elseif spell == GetSpellInfo(FS.spellID["BRG"]) then
		self.cooldowns["BRG"] = self.now + 20
	-- Explosive Trap 
	elseif spell == GetSpellInfo(FS.spellID["ET"]) then
		self.cooldowns["ET"] = self.now + 24
	-- Kill Shot
	elseif spell == GetSpellInfo(FS.spellID["KS"]) then
		self.cooldowns["KS"] = self.now + 10
	-- Serpent Sting
	-- Not really a CD but we don't want to fire it too often unnecessarily
	elseif spell == GetSpellInfo(FS.spellID["AS"]) then
		self.cooldowns["SpS"] = self.now + 15
	end
end

-- calculate focus when casting given spell
local function CalcFocus(self, parent)
	local spell = self.spell

	-- Explosive Shot
	if spell == GetSpellInfo(FS.spellID["ES"]) then
		if self.lnl then
			self.lnl = nil
		else
			self.focus = self.focus - self.cost["ES"]
		end
	-- Kill Command
	elseif spell == GetSpellInfo(FS.spellID["KC"]) then
			self.focus = self.focus - self.cost["KC"]
	-- Chimera Shot
	elseif spell == GetSpellInfo(FS.spellID["CS"]) then
			self.focus = self.focus - self.cost["CS"]
	-- Black Arrow
	elseif spell == GetSpellInfo(FS.spellID["BA"]) then
			self.focus = self.focus - self.cost["BA"]
	-- A Murder of Crows
	elseif spell == GetSpellInfo(FS.spellID["AMoC"]) then
			self.focus = self.focus - self.cost["AMoC"]
	-- Arcane Shot
	elseif spell == GetSpellInfo(FS.spellID["AS"]) then
		if self.buffs["TotH"] < self.now then
			self.focus = self.focus - 30
		else
			self.focus = self.focus - 10
		end
	-- Glaive Toss
	elseif spell == GetSpellInfo(FS.spellID["GT"]) then
			self.focus = self.focus - self.cost["GT"]
	-- Powershot
	elseif spell == GetSpellInfo(FS.spellID["PS"]) then
			self.focus = self.focus - self.cost["PS"]
	-- Barrage
	elseif spell == GetSpellInfo(FS.spellID["BRG"]) then
			self.focus = self.focus - self.cost["BRG"]
	-- Aimed Shot
	elseif spell == GetSpellInfo(FS.spellID["AimS"]) then
		--if parent ~= "fs" then
		if self.buffs["TotH"] < self.now then
			self.focus = self.focus - 50
		else
			self.focus = self.focus - 30
		end
	-- Steady, Cobra and Focus Shot focus gains
	elseif parent ~= "fs" and (spell == GetSpellInfo(FS.spellID["CoS"]) or spell == GetSpellInfo(FS.spellID["SS"]) or spell == GetSpellInfo(FS.spellID["FS"])) then
		self.focus = self.focus + self.gains
	end
end

-- returns focus after a given time to regenerate at a given startvalue
local function CalcReg(self, timeToReg, regenRate)
	self.focus = self.focus + regenRate * timeToReg * self.haste

	if FS.spec == "BM" and self.focus > 120 then self.focus = 120 end
	if FS.spec == "SV" and self.focus > 100 then self.focus = 100 end
	
	if FS.spec == "MM" and FS.perks then 
		if self.focus > 120 then self.focus = 120 end
	else
		if self.focus > 100 then self.focus = 100 end
	end
end

-- stabilize recommendation by making it harder to switch to a different spell once it has been determined at castSent
local function Damping(self, acrn, minFocus, maxFocus)
	if not self.recom or FS.state == "castSent" then
		return minFocus, maxFocus
	elseif self.recom == GetSpellInfo(FS.spellID[acrn]) then
		if minFocus and minFocus > 0 then minFocus = minFocus - FS.db.profile.damping end
		if maxFocus and maxFocus < 100 then maxFocus = maxFocus + FS.db.profile.damping end
	else
		if minFocus and minFocus > 0 then minFocus = minFocus + FS.db.profile.damping end
		if maxFocus and maxFocus < 100 then maxFocus = maxFocus - FS.db.profile.damping end
	end
	return minFocus, maxFocus
end

-- return the amount of focus that is regenerated (w/o SS/CoS/FS) until given spell is off CD again or debuff run off
local function RegTilRdy(self, acrn)
	local dur
	if self.cooldowns[acrn] and self.cooldowns[acrn] >= self.now then dur = self.cooldowns[acrn] - self.now
	elseif self.debuffs[acrn] and self.debuffs[acrn][FS.tarGUID] and self.debuffs[acrn][FS.tarGUID] >= self.now then dur = self.debuffs[acrn][FS.tarGUID] - self.now
	else return 0
	end
	return ( self.regenRate * self.haste * dur)
end

-- return true if player would be overcapped after num SS/CoS or cost free shots
local function Overcapped(self, isReg, num)
	local regen, ctime, cap
	
	if isReg then
	-- if either SS or CoS is used add the gains
		if FS.talents["FS"] then ctime = 2.5 / self.haste 
		else ctime = 2 / self.haste end-- Same 2.0 seconds for both SS and CoS
		regen = ctime * self.haste * self.regenRate + self.gains
	else regen = self.regenRate * self.haste
	end
	
	-- if num is omitted assume 1
	if not num then num = 1 end
	regen = num * regen
	
	if FS.spec == "BM" then cap = 120 end
	if FS.spec == "SV" then cap = 100 end
	if FS.spec == "MM" and FS.perks then cap = 120 else cap = 100 end

	if (self.focus + regen) > cap then return true
	else return false
	end
end

-- returns true if the last shot was the given one
local function IsLastShot(self, spell)
	if self.spellName[0] and self.spellName[0] == spell then return true
	elseif self.spellName[1] and self.spellName[1] == spell then return true
	elseif self.spellName[2] and self.spellName[2] == spell then return true
	else return false
	end
end

-- returns true if the next to last shot was the given one
local function IsNextToLastShot(self, spell)
	if self.spellName[0] and self.spellName[1] == spell then return true
	elseif self.spellName[1] and self.spellName[2] == spell then return true
	elseif self.spellName[2] and self.spellName[3] == spell then return true
	else return false
	end
end

-- return true if next shot should be a SS or CoS to get an SF proc (with or without overcap prevention)
-- Experimentation shows that if Focusing Shot and Steady Focus talents are enabled, then a single FS will provide the Steady Focus buff
local function CheckSF(self)
	if FS.talents["FS"] then return false --Focusing Shot
	else 
		local steady = GetSpellInfo(FS.spellID["SS"])
		local cobra = GetSpellInfo(FS.spellID["CoS"])
	
		if FS.talents["SF"] and
			(self:IsLastShot(steady) and not self:IsNextToLastShot(steady)) or
			(self:IsLastShot(cobra) and not self:IsNextToLastShot(cobra))
		then return true

		elseif self.buffs["SF"] >= self.now and self.buffs["SF"] <= (self.now + 4) then return true
		else return false
		end
	end
end

-- returns true if ES is off CD
local function CheckES(self, ...)
	local minFocus, maxFocus = ...
	if self.buffs["LnL"] > self.now then minFocus = 0
	else minFocus = self.cost["ES"]
	end
	if not maxFocus then maxFocus = 120 end
	minFocus, maxFocus = self:Damping("ES", minFocus, maxFocus)
	return self.cooldowns["ES"] <= (self.now + (FS.db.profile.maxWait / 1000)) and
		self.focus >= minFocus and
		self.focus <= maxFocus
end

-- returns true if CS is off CD and focus meets given requirements
local function CheckCS(self, ...)
	local minFocus, maxFocus = ...
	if not minFocus then minFocus = self.cost["CS"] end
	if not maxFocus then maxFocus = 120 end
	minFocus, maxFocus = self:Damping("CS", minFocus, maxFocus)
	return self.cooldowns["CS"] <= (self.now + (FS.db.profile.maxWait / 1000)) and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if AMoC is off CD and focus meets given requirements
local function CheckAMoC(self, ...)
	local FS_TTL = FS:GetModule("TimeToLife", true)
	if not FS_TTL:IsEnabled() then FS_TTL = false end
	if FS_TTL and FS_TTL.delta < FS.db.profile.AMoC.minTTL then return false end
	
	local minFocus, maxFocus = ...
	if not minFocus then minFocus = self.cost["AMoC"] end
	if not maxFocus then maxFocus = 120 end
	minFocus, maxFocus = self:Damping("AMoC", minFocus, maxFocus)
	
	if FS.db.profile.AMoC.suggestAMoC == "never" then return false
	elseif (FS.db.profile.AMoC.suggestAMoC == "level" and UnitLevel("target") ~= -1) then return false
	else return (FS.talents["AMoC"] and
		self.debuffs["AMoC"][FS.tarGUID] <= self.now and
		self.cooldowns["AMoC"] <= self.now and
		self.focus >= minFocus and self.focus <= maxFocus)
	end
end

-- returns true if Serpent Sting is not applied or if it will expire in the next 3 seconds
local function CheckSpS(self, ...)
	
	local minFocus, maxFocus = ...
	if not minFocus then minFocus = self.cost["AS"] end
	if not maxFocus then maxFocus = 120 end
	minFocus, maxFocus = self:Damping("SpS", minFocus, maxFocus)
	
	return (self.debuffs["SpS"][FS.tarGUID] <= self.now) and
	(self.focus >= minFocus and self.focus <= maxFocus) and
	(self.cooldowns["SpS"] <= (self.now+3))

end

-- returns true if AimS meets focus requirement
local function CheckAimS(self, ...)
	local minFocus, maxFocus, CarefulAim = ...
	if not maxFocus then maxFocus = 120 end
	if not minFocus then minFocus = self.cost["AimS"]
	else minFocus = minFocus + self.cost["AimS"] end
	if not CarefulAim then CarefulAim = false end

	--Save focus for BRG if enabled and its CD is less than 4 sec and not in CA phase
	if FS.talents["BRG"] and not CarefulAim and
		self.cooldowns["BRG"] < (self.now + 4) and
		((FS.db.profile.Brg.suggestBrg == "always") or (FS.db.profile.Brg.suggestBrg == "level" and UnitLevel("target") == -1))
		
	then
		--FS:Print("Building for Brg")
		minFocus = minFocus + self.cost["BRG"]
	end

	minFocus, maxFocus = self:Damping("AimS", minFocus, maxFocus)
	return self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if focus requirements are met
local function CheckAS(self, ...)

	local minFocus, maxFocus = ...
	if not maxFocus then maxFocus = 120 end
	if not minFocus then minFocus = self.cost["AS"]
	else minFocus = minFocus + self.cost["AS"] end

	--Save focus for BRG if enabled and its CD is less than 5 sec
	if FS.talents["BRG"] and
		self.cooldowns["BRG"] < (self.now + 5) and
		((FS.db.profile.Brg.suggestBrg == "always") or (FS.db.profile.Brg.suggestBrg == "level" and UnitLevel("target") == -1))
	then
		minFocus = minFocus + self.cost["BRG"] 
	end

	minFocus, maxFocus = self:Damping("AS", minFocus, maxFocus)
	return self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if BA is off CD and focus meets given requirements
local function CheckBA(self, ...)
	if FS.db.profile.ba.suggestBA == "never" or (FS.db.profile.ba.suggestBA == "level" and UnitLevel("target") ~= -1) then return false end
	local FS_TTL = FS:GetModule("TimeToLife", true)
	if not FS_TTL:IsEnabled() then FS_TTL = false end
	if FS_TTL and FS_TTL.delta < FS.db.profile.ba.minTTLBA then return false end
	local minFocus, maxFocus = ...
	if not minFocus or not FS.db.profile.ba.saveFocusBA then minFocus = self.cost["BA"] end
	if not maxFocus then maxFocus = 120 end
	minFocus, maxFocus = self:Damping("BA", minFocus, maxFocus)
	return self.debuffs["BA"][FS.tarGUID] <= self.now and self.cooldowns["BA"] <= self.now and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if focus meets given requirements
local function CheckET(self, ...)
	if FS.db.profile.ET.suggestET == "never" or (FS.db.profile.ET.suggestET == "level" and UnitLevel("target") ~= -1) then return false end
	
	local FS_TTL = FS:GetModule("TimeToLife", true)
	if not FS_TTL:IsEnabled() then FS_TTL = false end
	if FS_TTL and FS_TTL.delta < FS.db.profile.ET.minTTL then return false end
	
	local minFocus, maxFocus = ...
	if not minFocus then minFocus = 0 end
	if not maxFocus then maxFocus = 120 end
	minFocus, maxFocus = self:Damping("ET", minFocus, maxFocus)
	return self.cooldowns["ET"] <= self.now and self.focus >= minFocus and self.focus <= maxFocus
end

-- returns true if KS is off CD and focus meets given requirements
local function CheckKS(self, ...)
	if FS.ksslot == 0 then return false end
	
	local maxFocus = ...
	if not maxFocus then maxFocus = 120 end
	_, maxFocus = self:Damping("KS", nil, maxFocus)

	local targetHealthOK
	if FS.perks and FS.spec == "MM" then targetHealthOK = self.tarHP <= 0.35
	else targetHealthOK = self.tarHP <= 0.2 end

	if self.cooldowns["KS"] <= (self.now + (FS.db.profile.maxWait / 1000)) and targetHealthOK then return true
	else return false
	end
	--FS:Print(" CheckKS Now: "..FS:Round(self.now,1).."  CD: "..FS:Round(self.cooldowns["KS"],1))
end

-- returns true if focus requirements are met
local function CheckGT(self, ...)
	local minFocus, maxFocus = ...
		
	if not minFocus then minFocus = self.cost["GT"] end
	if not maxFocus then maxFocus = 120 end
	minFocus, maxFocus = self:Damping("GT", minFocus, maxFocus)
	
	if (FS.db.profile.GT.suggestGT == "level" and UnitLevel("target") ~= -1) then return false
	elseif FS.db.profile.GT.suggestGT == "never" then return false
	else return FS.talents["GT"] and self.cooldowns["GT"] <= self.now and self.focus >= minFocus and self.focus <= maxFocus
	end
end

-- returns true if focus requirements are met
local function CheckPS(self, ...)
	local minFocus, maxFocus = ...
	if not minFocus then minFocus = self.cost["PS"] end
	if not maxFocus then maxFocus = 120 end
	minFocus, maxFocus = self:Damping("PS", minFocus, maxFocus)
	
	if FS.db.profile.PS.suggestPS == "never" then return false
	elseif (FS.db.profile.PS.suggestPS == "level" and UnitLevel("target") ~= -1) then return false
	else return FS.talents["PS"] and self.cooldowns["PS"] <= self.now and self.focus >= minFocus and self.focus <= maxFocus
	end
end

-- returns true if focus requirements are met
local function CheckBRG(self, ...)
	local minFocus, maxFocus = ...
	if not minFocus then minFocus = self.cost["BRG"] end
	if not maxFocus then maxFocus = 120 end
	minFocus, maxFocus = self:Damping("BRG", minFocus, maxFocus)
	
	if FS.db.profile.Brg.suggestBrg == "never" then return false
	elseif (FS.db.profile.Brg.suggestBrg == "level" and UnitLevel("target") ~= -1) then return false
	else return FS.talents["BRG"] and self.cooldowns["BRG"] <= self.now and self.focus >= minFocus and self.focus <= maxFocus
	end

end

-- returns true if KC is off CD and focus meets given requirements
local function CheckKC(self, ...)
	local minFocus, maxFocus = ...
	if not UnitExists("pet") or UnitIsDead("pet") then return false end
	minFocus = self.cost["KC"]
	if not maxFocus then maxFocus = 120 end
	minFocus, maxFocus = self:Damping("KC", minFocus, maxFocus)
	return self.cooldowns["KC"] <= (self.now + (FS.db.profile.maxWait / 1000)) and self.focus >= minFocus and self.focus <= maxFocus
end

local function CheckFZ(self)
	if self.buffs["FZ"] == 5 then
		self.buffs["FZ"] = 0
		return true
	else
		return false
	end
end

-- returns true if casting doesn't lead to overcap
local function CheckCoS(self, ...)
	if not FS.talents["FS"] then return not self:Overcapped(true)
	else return false end
end

local function CheckSS(self, ...)
	if not FS.talents["FS"] then return not self:Overcapped(true)
	else return false end
end

local function CheckFS(self, ...)
	if FS.talents["FS"] then return not self:Overcapped(true)
	else return false end
end

local function CheckDB(self, ...)
	local FS_TTL = FS:GetModule("TimeToLife", true)
	if not FS_TTL:IsEnabled() then FS_TTL = false end
	if FS_TTL and FS_TTL.delta < FS.db.profile.db.minTTL then return false end

	if FS.db.profile.db.suggestDB == "never" then return false
	elseif (FS.db.profile.db.suggestDB == "level" and UnitLevel("target") ~= -1) then return false
	else return FS.talents["DB"] and self.cooldowns["DB"] <= self.now
	end
end



-- pick the most beneficial spell based on given resources
local function PickSpell(self, btnName)
	local focus, spell = self.focus

	self:GetSpellCosts()
	self.recom = FS[btnName].recom

	-- SV rotation
	if FS.spec == "SV" then
		-- Explosive Shot
		if self:CheckES() then
			spell = FS.spellID["ES"]
		-- Cobra Shot for Steady Focus
		elseif self:CheckSF(true) then
			spell = FS.spellID["CoS"]
		-- Explosive Trap
		elseif self:CheckET() then
			spell = FS.spellID["ET"]
		-- Black Arrow
		elseif self:CheckBA((self.cost["ES"] + self.cost["BA"] - self:RegTilRdy("ES"))) then
			spell = FS.spellID["BA"]
		-- Serpent Sting
		elseif self:CheckSpS() then
			spell = FS.spellID["AS"]
		-- Barrage
		elseif self:CheckBRG() then
			spell = FS.spellID["BRG"]
		-- A Murder of Crows
		elseif self:CheckAMoC() then
			spell = FS.spellID["AMoC"]
		-- Dire Beast
		elseif self:CheckDB() then
			spell = FS.spellID["DB"]
		-- Glaive Toss
		elseif self:CheckGT() then
			spell = FS.spellID["GT"]
		-- Powershot
		elseif self:CheckPS() then
			spell = FS.spellID["PS"]
		-- Arcane Shot
		elseif self:CheckAS((self.cost["ES"] + self.cost["BA"] - self:RegTilRdy("ES"))) then
			spell = FS.spellID["AS"]
		-- Cobra Shot
		elseif self:CheckCoS() then	
			spell = FS.spellID["CoS"]
		-- Focusing Shot
		elseif self:CheckFS() then	
			spell = FS.spellID["FS"]
		else
			spell = FS.spellID["AS"] -- Focus Dump
		end
	-- BM rotation
	elseif FS.spec == "BM" then
		-- Kill Shot
		if self:CheckKS() then
			spell = FS.spellID["KS"]
		-- Frenzy
		elseif self:CheckFZ() then
			spell = FS.spellID["FZ"]
		-- Cobra Shot for Steady Focus
		elseif self:CheckSF(true) then
			spell = FS.spellID["CoS"]
		-- Kill Command
		elseif self:CheckKC() then
			spell = FS.spellID["KC"]
		-- Barrage
		elseif self:CheckBRG() then
			spell = FS.spellID["BRG"]
		-- A Murder of Crows
		elseif self:CheckAMoC() then
			spell = FS.spellID["AMoC"]
		-- Dire Beast
		elseif self:CheckDB(true) then
			spell = FS.spellID["DB"]
		-- Explosive Trap
		elseif self:CheckET() then
			spell = FS.spellID["ET"]
		-- Glaive Toss
		elseif self:CheckGT() then
			spell = FS.spellID["GT"]
		-- Powershot
		elseif self:CheckPS() then
			spell = FS.spellID["PS"]
		-- Arcane Shot
		elseif self:CheckAS((self.cost["KC"] - self:RegTilRdy("KC"))) then
			spell = FS.spellID["AS"]
		-- Focusing Shot
		elseif self:CheckFS() then	
			spell = FS.spellID["FS"]
		-- Cobra Shot
		else
			spell = FS.spellID["AS"]
		end
	-- MM rotation
	elseif FS.spec == "MM" then
		-- TODO - Perhaps try to predict with TTL if target will be below 80% health when shot is sent...
		local CA = self.buffs["RF"] > self.now or self.tarHP > 0.8
		if FS.db.profile.CA.suggestCA == "never" then CA = false 
		elseif (FS.db.profile.CA.suggestCA == "level" and UnitLevel("target") ~= -1) then CA = false end
		-- Kill Shot
		if self:CheckKS() then
			spell = FS.spellID["KS"]
		-- Chimaera Shot
		elseif self:CheckCS() then
			spell = FS.spellID["CS"]
		-- Steady Focus Shot
		elseif self:CheckSF(true) then
			spell = FS.spellID["SS"]
		-- Barrage
		elseif self:CheckBRG() and not CA then
			spell = FS.spellID["BRG"]
		-- A Murder of Crows
		elseif self:CheckAMoC() and not CA  then
			spell = FS.spellID["AMoC"]
		-- Dire Beast
		elseif self:CheckDB(true)  and not CA then
			spell = FS.spellID["DB"]
		-- Explosive Trap
		elseif self:CheckET()  and not CA then
			spell = FS.spellID["ET"]
		-- Glaive Toss
		elseif self:CheckGT()  and not CA then
			spell = FS.spellID["GT"]
		-- Powershot
		elseif self:CheckPS()  and not CA then
			spell = FS.spellID["PS"]
		-- Aimed Shot
		elseif self:CheckAimS((self.cost["CS"] - self:RegTilRdy("CS")),120,CA) then
			spell = FS.spellID["AimS"]
		-- Focusing Shot
		elseif self:CheckFS() then	
			spell = FS.spellID["FS"]
		-- Steady Shot
		elseif self:CheckSS(true) then
			spell = FS.spellID["SS"]
		else
			spell = FS.spellID["AimS"] --Default
		end
	-- no spec
	else
		-- Kill Shot
		if self:CheckKS() then
			spell = FS.spellID["KS"]
		-- Explosive Trap
		elseif self:CheckET() then
			spell = FS.spellID["ET"]
		-- Arcane Shot
		elseif self:CheckAS((self.cost["AS"])) then
			spell = FS.spellID["AS"]
		-- Cobra Shot
		elseif FS.level > 80 then	
			spell = FS.spellID["CoS"]
		-- Steady Shot
		else
			spell = FS.spellID["SS"]
		end
	end

	local spellName = GetSpellInfo(spell)
	
	-- draw the corresponding texture
	if self.spellName[0] then
		table.insert(self.spellName, 1, self.spellName[0])
		table.remove(self.spellName, 5)
		table.insert(self.spellSentAt, 1, self.spellSentAt[0])
		table.remove(self.spellSentAt, 5)
		table.insert(self.spellStartedAt, 1, self.spellStartedAt[0])
		table.remove(self.spellStartedAt, 5)
		table.insert(self.spellEndedAt, 1, self.spellEndedAt[0])
		table.remove(self.spellEndedAt, 5)
		self.spellName[0] = nil
		self.spellSentAt[0] = nil
		self.spellStartedAt[0] = nil
		self.spellEndedAt[0] = nil
	end
	
	if self.spellName[1] then
		table.insert(self.spellName, 1, spellName)
		table.remove(self.spellName, 5)
		table.insert(self.spellSentAt, 1, self.now)
		table.remove(self.spellSentAt, 5)
		table.insert(self.spellStartedAt, 1, self.now + FS.avrgLatency)
		table.remove(self.spellStartedAt, 5)
		table.insert(self.spellEndedAt, 1, self.now + FS:GetCastDelta(spellName, self.haste)) --> new self.now for next button - ToDo: get it as precise as possible
		table.remove(self.spellEndedAt, 5)
	else
		self.spellName[1] = spellName
		self.spellSentAt[1] = self.now
		self.spellStartedAt[1] = self.now + FS.avrgLatency
		self.spellEndedAt[1] = self.now + FS:GetCastDelta(spellName, self.haste) --> new self.now for next button - ToDo: get it as precise as possible
	end
	FS[btnName].btn.texture:SetTexture(GetSpellTexture(spell))
	--FS[btnName]:DrawCDF(FS:GetGCD())
	FS[btnName].recom = GetSpellInfo(spell)
	local inRange = IsSpellInRange(GetSpellInfo(FS.spellID["AutoS"]), "target")
	
	if inRange == 1 then FS[btnName].btn.texture:SetVertexColor(1, 1, 1)
	else FS[btnName].btn.texture:SetVertexColor(1, 0.15, 0.15)
	end
end

-- creates a resource object 
function FS:CreateResources()
	local obj = {
		now = 0,
		focus = 100,
		es = 0,
		haste = 1,
		tarHP = 0,
		regenRate = 4,
		spell = nil,
		spellName = {},
		spellSentAt = {},
		spellStartedAt = {},
		spellEndedAt = {},
		kscd = false, --KillShot Cooldown reset
		charges = {
			["LnL"] = 0,
		},
		buffs = {
			["LnL"] = 0,
			["SF"] = 0,
			["TotH"] = 0,
			["BW"] = 0,
			["RF"] = 0,
			["FZ"] = 0,
			["TT"] = 0, --Telaari Talbuk
			["FWW"] = 0, --Frostwolf
		},
		debuffs = {
			["BA"] = {},
			["AMoC"] = {},
			["SpS"] = {},
		},
		cooldowns = {
			["CS"] = 0,
			["BA"] = 0,
			["ES"] = 0,
			["KS"] = 0,
			["AimS"] = 0,
			["KC"] = 0,
			["ET"] = 0,
			["DB"] = 0,
			["AMoC"] = 0,
			["GT"] = 0,
			["PS"] = 0,
			["BRG"] = 0,
			["SpS"] = 0,
		},
		focusGains = {
			["CoS"] = 0,
			["SS"] = 0,
			["FS"] = 0,
		},
		lastCast = {
			["CS"] = 0,
			["BA"] = 0,
			["ES"] = 0,
			["KS"] = 0,
			["AimS"] = 0,
			["KC"] = 0,
			["ET"] = 0,
			["CoS"] = 0,
			["SS"] = 0,
			["PS"] = 0,
			["BRG"] = 0,
		},
		cost = {
			["ES"] = 15,
			["KC"] = 40,
			["CS"] = 35,
			["BA"] = 35,
			["AS"] = 30,
			["AimS"] = 50,
			["AMoC"] = 30,
			["GT"] = 15,
			["PS"] = 15,
			["BRG"] = 60,
		},
		CheckES = CheckES,
		CheckCS = CheckCS,
		CheckBA = CheckBA,
		CheckAMoC = CheckAMoC,
		CheckDB = CheckDB,
		CheckET = CheckET,
		CheckKS = CheckKS,
		CheckAimS = CheckAimS,
		CheckGT = CheckGT,
		CheckPS = CheckPS,
		CheckBRG = CheckBRG,
		CheckAS = CheckAS,
		CheckKC = CheckKC,
		CheckCoS = CheckCoS,
		CheckSF = CheckSF,
		CheckFZ = CheckFZ,
		CheckFS = CheckFS,
		CheckSS = CheckSS,
		CheckSpS = CheckSpS,
		IsLastShot = IsLastShot,
		IsNextToLastShot = IsNextToLastShot,
		Damping = Damping,
		RegTilRdy = RegTilRdy,
		Overcapped = Overcapped,
		PickSpell = PickSpell,
		GetResources = GetResources,
		SetResources = SetResources,
		FocusGains = FocusGains,
		GetBuffs = GetBuffs,
		GetDebuffs = GetDebuffs,
		PurgeDebuffs = PurgeDebuffs,
		CastSpell = CastSpell,
		CalcTimers = CalcTimers,
		CalcFocus = CalcFocus,
		CalcReg = CalcReg,
		GetSpellCosts = GetSpellCosts,
		Round = Round,
	}
	return obj
end

------------
--- button ---
------------

-- draw the cooldown frame or animate the frame in some style
local function DrawCDF(self, start, duration)
	local now = GetTime()
	if duation == 0 then duration = 1 end
	--local FS_ANI = FS:GetModule("Animation", true)
	--if FS_ANI:IsEnabled() and FS.db.profile.animation.style ~= "none" then return false end
	if FS.isInGCD ~= (start + duration) and (start + duration) > 0 then
		--FS:Print("DrawCDF: Start: "..start.." Duration: "..duration)
		self.cdf:SetCooldown(GetTime(), duration)
		self.cdf:Show()
		FS.isInGCD = start + duration
	elseif FS.isInGCD and FS.isInGCD > now and (start + duration) == 0 then
		self.cdf:Hide()
		FS.isInGCD = false
	elseif FS.isInGCD and (start + duration) == 0 then
		FS.isInGCD = false
	end
end

-- hides the objects frame
local function Hide(self)
	self.btn:Hide()
	self.cdf:Hide()
end

-- shows the objects frame
local function Show(self)
	self.btn:Show()
	if FS.isInGCD then self.cdf:Show() end
end

-- creates an button object
function FS:CreateButton(name)
	local obj = {
		name = name,
		resources = self:CreateResources(),
		mov = CreateFrame("Frame", name.."Mover", UIParent),
		btn = CreateFrame("Button", name.."Button", mov),
		cdf = CreateFrame("Cooldown", name.."CooldownFrame", btn, "CooldownFrameTemplate"),
		PickSpell = PickSpell,
		DrawCDF = DrawCDF,
		Hide = Hide,
		Show = Show,
	}

	obj.mov:SetFrameStrata("BACKGROUND")
	obj.mov:SetWidth(40)
	obj.mov:SetHeight(40)
	obj.mov:SetScale(self.db.profile[name]["scale"]/100)
	obj.mov:SetPoint(self.db.profile[name]["point"], self.db.profile[name]["x"], self.db.profile[name]["y"])
	obj.mov:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 32,})
	obj.mov:SetBackdropColor(0, 0, 0, 0.4)
	obj.mov:SetScript("OnMouseDown", function(self) self:StartMoving() end)
	obj.mov:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing() FS.db.profile[name]["point"], _, _, FS.db.profile[name]["x"], FS.db.profile[name]["y"] = self:GetPoint() end)
	obj.mov:SetMovable(true)
	obj.mov:SetClampedToScreen(true)
	obj.mov.str = obj.mov:CreateFontString("String", "ARTWORK", "GameFontNormal")
	obj.mov.str:SetFont(GameFontNormal:GetFont(), 10)
	obj.mov.str:SetJustifyH("CENTER")
	obj.mov.str:SetJustifyV("TOP")
	obj.mov.str:SetPoint("CENTER", obj.mov, "CENTER", 0, 30)
	obj.mov.str:SetTextColor(1, 1, 1, 1)
	obj.mov.str:SetText(name)

	-- hide anchor frame if it is locked
	if self.db.profile.lock then
		obj.mov:SetAlpha(0)
		obj.mov:EnableMouse(false)
	-- otherwise show and make it moveable
	else
		obj.mov:SetAlpha(1)
		obj.mov:EnableMouse(true)
	end
	obj.btn:SetWidth(32)
	obj.btn:SetHeight(32)
	obj.btn:EnableMouse(false)
	obj.btn:SetAlpha(self.db.profile[name]["alpha"]/100)
	obj.btn:SetScale(self.db.profile[name]["scale"]/100)
	obj.btn:SetPoint("TOPLEFT", obj.mov)
	obj.btn.texture = obj.btn:CreateTexture(nil, "BACKGROUND")
	obj.btn.texture:SetTexture(nil)
	obj.btn.texture:ClearAllPoints()
	obj.btn.texture:SetAllPoints(obj.btn)
	obj.cdf:SetHeight(32)
	obj.cdf:SetWidth(32)
	obj.cdf:ClearAllPoints()
	obj.cdf:SetAllPoints(obj.btn)
	obj.cdf:SetAlpha(1)

	return obj
end

-------------
--- Debug ---
-------------

-- makes sure that the given value is concatenable
function FS:mkStr(val, round, modula)
	local out = val
	if val == nil then out = "nil"
	elseif val == false then out = "false"
	elseif type(val) ~= "number" and type(val) ~= "string" then out = type(val)
	end
	if modula and tonumber(out) then out = val % modula end
	if round and tonumber(out) then out = FS:Round(out, round) end
	return out
end

-- takes a snapshot of current resources and logs them
function FS:GetLogEntry(name)
	local entry = {
		name = name,
		recom = FS.button1.recom,
		state = FS.state,
		times = "real now: "..GetTime().." / self.now: "..FS:Round(FS.button1.resources.now, 3).." / lastCastSentAt: "..self:mkStr(FS.lastCastSentAt).." / lastCastStartedAt: "..self:mkStr(FS.lastCastStartedAt).." / lastCastSucceededAt: "..self:mkStr(FS.lastCastSucceededAt),
		focus = "real focus: "..UnitPower("player").." / self.focus: "..FS:Round(FS.button1.resources.focus, 3),
		spellNames = "sNam[0]: "..self:mkStr(FS.spellName[0]).." / sNam[1]: "..self:mkStr(FS.spellName[1]).." / sNam[2]: "..self:mkStr(FS.spellName[2]).." / sNam[3]: "..self:mkStr(FS.spellName[3]).." / sNam[4]: "..self:mkStr(FS.spellName[4]),
		spellSentAt = "sSend[0]: "..self:mkStr(FS.spellSentAt[0], 3, 100).." / sSend[1]: "..self:mkStr(FS.spellSentAt[1], 3, 100).." / sSend[2]: "..self:mkStr(FS.spellSentAt[2], 3, 100).." / sSend[3]: "..self:mkStr(FS.spellSentAt[3], 3, 100).." / sSend[4]: "..self:mkStr(FS.spellSentAt[4], 3, 100),
		spellStartedAt = "sStart[0]: "..self:mkStr(FS.spellStartedAt[0], 3, 100).." / sStart[1]: "..self:mkStr(FS.spellStartedAt[1], 3, 100).." / sStart[2]: "..self:mkStr(FS.spellStartedAt[2], 3, 100).." / sStart[3]: "..self:mkStr(FS.spellStartedAt[3], 3, 100).." / sStart[4]: "..self:mkStr(FS.spellStartedAt[4], 3, 100),
		spellEndedAt = "sEnd[0]: "..self:mkStr(FS.spellEndedAt[0], 3, 100).." / sEnd[1]: "..self:mkStr(FS.spellEndedAt[1], 3, 100).." / sEnd[2]: "..self:mkStr(FS.spellEndedAt[2], 3, 100).." / sEnd[3]: "..self:mkStr(FS.spellEndedAt[3], 3, 100).." / sEnd[4]: "..self:mkStr(FS.spellEndedAt[4], 3, 100),
		misc = "gcd: "..FS.gcd.." / inGCD: "..self:mkStr(FS.inGCD).." / es: "..self:mkStr(FS.button1.resources.es).." / haste = "..self:mkStr(FS.button1.resources.haste),
		regTilRdy = "CS: "..FS.button1.resources:RegTilRdy("CS").." / KC: "..FS.button1.resources:RegTilRdy("KC").." / ES: "..FS.button1.resources:RegTilRdy("ES"),
		charLnL = FS.button1.resources.charges["LnL"],
		debuffs = "",
		buffs = "SF: "..FS.button1.resources.buffs["SF"],
		cooldowns = "CS: "..FS.button1.resources.cooldowns["CS"].." / KC: "..FS.button1.resources.cooldowns["KC"].." / ES: "..FS.button1.resources.cooldowns["ES"].." / ET: "..FS.button1.resources.cooldowns["ET"].." / BA: "..FS.button1.resources.cooldowns["BA"],
		focusGains = "SS: "..FS.resources.focusGains["SS"].." / CoS: "..FS.resources.focusGains["CoS"],
		debuffCheck = FS.debuffCheck,
	}
	if not FS.db.profile.log then FS.db.profile.log = {} end
	table.insert(FS.db.profile.log, entry)
end

-- writes the snapshot to the savedvariables
function FS:WriteLogEntry(name)
	if type(self.log) ~= "table" then self.log = { lastSpell, lastState } end
	if FS.button1.recom ~= self.log.lastSpell or FS.state ~= self.log.lastState then
		self:GetLogEntry(name)
		if name == "After" then
			self.log.lastSpell = FS.button1.recom
			self.log.lastState = FS.state
		end
	end
end

-- creates the string for the debug frame
local function GetDebugStr(self)
	local str = self.str
	local i, a
	
	-- FS
	str.fs = "State: "..FS.state.."\n\n"
	for i = 0, 4 do
		str.fs = str.fs.."spell["..i.."]: "..FS:mkStr(FS.spellName[i]).." / "..FS:mkStr(FS.spellSentAt[i], 3, 100).." / "..FS:mkStr(FS.spellStartedAt[i], 3, 100).." / "..FS:mkStr(FS.spellEndedAt[i], 3, 100).."\n"
	end
	--str.fs = str.fs.."\nnow: "..FS:Round((FS.resources.now % 60), 2).."\n"
	str.fs = str.fs.."\nnow: "..FS:Round((FS.resources.now), 2).."\n"
	if not a then a = "" elseif a < 0 then a = 0 end
	str.fs = str.fs.."focus: "..UnitPower("player").."\n"
	if FS.isInGCD then str.fs = str.fs.."isInGCD: "..FS.isInGCD.."\n" else str.fs = str.fs.."isInGCD: false\n" end
	str.fs = str.fs.."latency: "..FS:Round(FS.avrgLatency, 3).."\n\n"
	str.fs = str.fs.."focusGains[SS]: "..FS.resources.focusGains["SS"].."\n"
	str.fs = str.fs.."focusGains[CoS]: "..FS.resources.focusGains["CoS"].."\n"
	-- button 1-2
	for i = 1, 2 do
		local n = "button"..i
		if FS.db.profile[n].enabled then
			local a
			--Recommended Shot
			str[n] = "Frame: "..n.."\n\n"
			if FS[n].resources.spell then str[n] = str[n].."recom: "..FS[n].recom.."\n"
			else str[n] = str[n].."recom:\n"
			end
			
			-- Time now
			if FS[n].resources.now then str[n] = str[n].."now: "..FS:Round((FS[n].resources.now ), 2).."\n"
			else str[n] = str[n].."now:\n"
			end
			
			if not a then a = "" elseif a < 0 then a = 0 end -- What is this for?
			-- Focus at end of shot (Current focus - Cost of shot + Regen)
			if FS[n].resources.focus then str[n] = str[n].."focus: "..FS:Round(FS[n].resources.focus, 2).."\n"
			else str[n] = str[n].."focus:\n"
			end
			
			if FS[n].resources.focus then str[n] = str[n].."focus: "..FS:Round(FS[n].resources.focus, 2).."\n"
			else str[n] = str[n].."focus:\n"
			end
			
		end
	end
end

-- writes the text at the frame
local function WriteText(self)
	local i
	self.btn.fs:SetPoint("TOPLEFT", self.btn, "TOPLEFT", 20, -15)
	self.btn.fs:SetText(self.str.fs)
	local width, height, num, x, y = 250, 220
	for i = 1, 2 do
		local n = "button"..i
		if FS.db.profile[n].enabled then
			num = i
			if i < 3 then x = i * width + 20 else x = (i - 2) * width + 20 end
			if i < 3 then y = -15 else y = -15 - height end
			self.btn[n]:SetPoint("TOPLEFT", self.btn, "TOPLEFT", x, y)
			self.btn[n]:SetText(self.str[n])
		end
	end
	if num == 1 then x = 2 * width + 20 else x = 3 * width + 20 end
	y = height - 15
	self.btn:SetWidth(x)
	self.btn:SetHeight(y)
end

-- create an frame object for debugging information
function FS:CreateDebugFrame()
	local obj = {
		btn = CreateFrame("Button", btn ,UIParent),
		str = {},
		GetDebugStr = GetDebugStr,
		WriteText = WriteText,
	}
	obj.btn:SetBackdrop({bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", tile = true, tileSize = 32, edgeSize = 32, insets = {left = 11, right = 12, top = 12, bottom = 11}})
	obj.btn:SetPoint(self.db.profile.debug.point, self.db.profile.debug.x, self.db.profile.debug.y)
	obj.btn:EnableMouse(true)
	obj.btn:SetMovable(true)
	obj.btn:SetScript("OnMouseDown", function(self) self:StartMoving() end)
	obj.btn:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing() FS.db.profile.debug.point, _, _, FS.db.profile.debug.x, FS.db.profile.debug.y = self:GetPoint() end)
	obj.btn.fs = obj.btn:CreateFontString("String", "ARTWORK", "GameFontNormal")
	obj.btn.fs:SetFont(GameFontNormal:GetFont(), 8)
	obj.btn.fs:SetTextColor(1, 1, 1, 1)
	obj.btn.fs:SetJustifyH("LEFT")
	obj.btn.fs:SetJustifyV("TOP")
	for i = 1, 2 do
		local n = "button"..i
		obj.btn[n] = obj.btn:CreateFontString("String", "ARTWORK", "GameFontNormal")
		obj.btn[n]:SetFont(GameFontNormal:GetFont(), 8)
		obj.btn[n]:SetTextColor(1, 1, 1, 1)
		obj.btn[n]:SetJustifyH("LEFT")
		obj.btn[n]:SetJustifyV("TOP")
	end
	obj:GetDebugStr()
	obj:WriteText()
	return obj
end